import gzip
import io
import logging
import sys
import typing
from urllib.request import urlopen

DEFAULT_LOG_FMT = '%(asctime)s %(name)-20s %(levelname)-3s : %(message)s'


def setup_logging(level: int = logging.INFO,
                  log_fmt: str = DEFAULT_LOG_FMT):
    """
        Create a basic configuration for the logging library. Set up console and file handler using provided `log_fmt`.
        :param level: verbosity to use, INFO by default
        :param log_fmt: format string for logging
        """
    # create logger
    logger = logging.getLogger()
    logger.setLevel(level)
    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(level)
    # create formatter
    formatter = logging.Formatter(log_fmt)
    # add formatter to ch
    ch.setFormatter(formatter)
    # add ch to logger
    logger.addHandler(ch)


def looks_like_url(file: str) -> bool:
    return file.startswith('http://') or file.startswith('https://')


def looks_gzipped(file: str) -> bool:
    return file.endswith('.gz')


def open_text_io_handle(fh: typing.Union[typing.IO, str],
                        timeout: int = 30,
                        encoding: str = None) -> typing.TextIO:
    """
    Open a `io.TextIO` file handle based on `fh`.

    :param fh: a `str` or `typing.IO` to read from. If `str`, then it should be a path to a local file or a URL
    of a remote resource. Either `http` or `https` protocols are supported. The content will be uncompressed on the fly
    if the file name ends with `.gz`. If `fh` is an IO wrapper, the function ensures we get a text wrapper that uses
    given encoding.
    :param timeout: timeout in seconds used when accessing a remote resource
    :param encoding: encoding used to decode the input or the system preferred encoding if unset.
    :return: the `io.TextIO` wrapper
    """
    logger = logging.getLogger('hpotk.util')
    logger.debug(f'Opening {fh}')
    if encoding is None:
        encoding = sys.getdefaultencoding()
        logger.debug(f'Using default encoding \'{encoding}\'')
    else:
        logger.debug(f'Using provided encoding \'{encoding}\'')

    if isinstance(fh, str):
        # Can be a path to local file or URL
        if looks_like_url(fh):
            logger.debug(f'Looks like a URL: {fh}')
            if not isinstance(timeout, int) or timeout <= 0:
                raise ValueError(f'If {fh} looks like URL then timeout {timeout} must be a positive `int`')
            logger.debug(f'Downloading with timeout={timeout}s')
            handle = urlopen(fh, timeout=timeout)
        else:
            logger.debug(f'Looks like a local file: {fh}')
            handle = open(fh, 'rb')

        if looks_gzipped(fh):
            logger.debug(f'Looks like a gzipped data, uncompressing on the fly')
            return gzip.open(handle, mode='rt', newline='', encoding=encoding)
        else:
            logger.debug(f'Looks like un-compressed data')
            return io.TextIOWrapper(handle, encoding=encoding)
    elif isinstance(fh, typing.IO):
        if isinstance(fh, typing.BinaryIO):
            logger.debug(f'Looks like a binary IO')
            return io.TextIOWrapper(fh, encoding=encoding)
        elif isinstance(fh, typing.TextIO):
            return fh
        else:
            raise ValueError(f'Unexpected type {type(fh)}')
    else:
        raise ValueError(f'Unexpected type {type(fh)}')



import bisect
import typing
import unittest

class ValueCache:

    # T = typing.TypeVar('T', bound=Comparable)

    def __init__(self, key=None):
        self._values: typing.List = []
        self._key = key

    def get_idx(self, value) -> int:
        if len(self._values) == 0:
            self._values.append(value)
            return 0
        else:
            idx = bisect.bisect_left(self._values, value, key=self._key)
            if idx != len(self._values) and self._values[idx] != value:
                self._values.insert(idx, value)
            return idx


class TestValueCache(unittest.TestCase):

    def test_bla(self):
        cache = ValueCache()
        a = cache.get_idx('a')
        b = cache.get_idx('b')
        c = cache.get_idx('c')
        d = cache.get_idx('a')
        e = cache.get_idx('b')
        e = cache.get_idx('c')



import typing

from hpotk.model import TermId, CURIE_OR_TERM_ID
from ._base import AnnotationReference, Ratio
from ._base import HpoDisease, HpoDiseaseAnnotation, HpoDiseases


class SimpleHpoDiseaseAnnotation(HpoDiseaseAnnotation):

    def __init__(self, identifier: TermId,
                 ratio: Ratio,
                 references: typing.List[AnnotationReference],
                 modifiers: typing.List[TermId]):
        self._id = identifier
        self._ratio = ratio
        self._refs = references
        self._modifiers = modifiers

    @property
    def identifier(self) -> TermId:
        return self._id

    @property
    def ratio(self) -> Ratio:
        return self._ratio

    @property
    def references(self) -> typing.List[AnnotationReference]:
        return self._refs

    @property
    def modifiers(self) -> typing.List[TermId]:
        return self._modifiers


class SimpleHpoDisease(HpoDisease):

    def __init__(self, identifier: TermId,
                 name: str,
                 annotations: typing.Collection[HpoDiseaseAnnotation],
                 modes_of_inheritance: typing.Collection[TermId]):
        self._id = identifier
        self._name = name
        self._annotations = annotations
        self._modes_of_inheritance = modes_of_inheritance

    @property
    def identifier(self) -> TermId:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def annotations(self) -> typing.Collection[HpoDiseaseAnnotation]:
        return self._annotations

    @property
    def modes_of_inheritance(self) -> typing.Collection[TermId]:
        return self._modes_of_inheritance


class SimpleHpoDiseases(HpoDiseases):

    def __init__(self, diseases: typing.Iterable[HpoDisease], version: str = None):
        self._diseases = {d.identifier: d for d in diseases}
        self._version = version

    @property
    def diseases(self) -> typing.Collection[HpoDisease]:
        return self._diseases.values()

    def __getitem__(self, item: CURIE_OR_TERM_ID) -> typing.Optional[HpoDisease]:
        if isinstance(item, TermId):
            pass
        elif isinstance(item, str):
            item = TermId.from_curie(item)
        else:
            raise ValueError(f'Expected a `str` or `TermId` but got {type(item)}')
        try:
            return self._diseases[item]
        except KeyError:
            return None

    def __len__(self) -> int:
        return len(self._diseases)

    @property
    def version(self) -> typing.Optional[str]:
        return self._version



import abc
import typing

from hpotk.model import ID, CURIE_OR_TERM_ID, TERM, MINIMAL_TERM, Versioned
from hpotk.graph import GraphAware


class MinimalOntology(typing.Generic[ID, MINIMAL_TERM], GraphAware[ID], Versioned, metaclass=abc.ABCMeta):
    """
    Minimal ontology has the ontology graph and `hpotk.base.model.MinimalTerm`s.
    """

    @property
    @abc.abstractmethod
    def term_ids(self) -> typing.Iterator[ID]:
        """
        :return: an iterator over `TermId`s of primary and obsolete ontology terms
        """
        pass

    @property
    @abc.abstractmethod
    def terms(self) -> typing.Iterator[MINIMAL_TERM]:
        """
        :return: an iterator over current `Term`s (*not* obsolete `Term`s)
        """
        pass

    @abc.abstractmethod
    def get_term(self, term_id: CURIE_OR_TERM_ID) -> typing.Optional[MINIMAL_TERM]:
        """
        Get the current `Term` for a term ID.

        :param term_id: a `TermId` or a CURIE `str` (e.g. 'HP:1234567') representing *current* or *obsolete* term
        :return: the current `Term` or `None` if the ontology does not contain the `TermId`
        """
        pass

    def __contains__(self, term_id: CURIE_OR_TERM_ID) -> bool:
        """
        Test if the ontology contains given `TermId`.

        If you are interested in processing the corresponding term, call `get_term(item)` instead.

        :param term_id: a `TermId` or a CURIE `str` (e.g. 'HP:1234567') representing *current* or *obsolete* term
        :return: `True` if the `TermId` is in the ontology and `False` otherwise
        """
        return self.get_term(term_id) is not None

    @abc.abstractmethod
    def __len__(self):
        """
        Get the number of the primary (non-obsolete) `Term`s in the ontology.

        :return: the number of primary `Term`s
        """
        pass


class Ontology(MinimalOntology[ID, TERM], metaclass=abc.ABCMeta):
    """
    An ontology with all information available for terms.
    """
    pass



import abc
import typing

from hpotk.algorithm import get_ancestors
from hpotk.model import Identified, TermId
from hpotk.ontology import MinimalOntology
from hpotk.constants.hpo.base import PHENOTYPIC_ABNORMALITY

from ._model import ValidationResult, ValidationResults, ValidationLevel, RuleValidator


class BaseOntologyRuleValidator(RuleValidator, metaclass=abc.ABCMeta):

    def __init__(self, ontology: MinimalOntology):
        self._ontology = ontology

    def _primary_term_id(self, identifier: TermId) -> typing.Optional[TermId]:
        """
        Map the provided `identifier` into the primary term ID in case the `identifier` is obsolete.
        """
        current_term = self._ontology.get_term(identifier)
        return current_term.identifier if current_term is not None else None


class AnnotationPropagationValidator(BaseOntologyRuleValidator):
    """
    Validator to check that a sequence of terms does not contain a term and its ancestor.

    The validator replaces obsolete term IDs with the current term IDs before performing the validation.
    """

    def __init__(self, ontology: MinimalOntology):
        super().__init__(ontology)

    def validate(self, items: typing.Sequence[Identified]) -> ValidationResults:
        term_ids = {self._primary_term_id(term.identifier) for term in items}
        results = []
        for term_id in term_ids:
            for ancestor in get_ancestors(self._ontology, source=term_id, include_source=False):
                if ancestor in term_ids:
                    current_term = self._ontology.get_term(term_id)
                    ancestor_term = self._ontology.get_term(ancestor)
                    results.append(
                        ValidationResult(level=ValidationLevel.ERROR,
                                         category='annotation_propagation',
                                         message=f'Terms should not contain both '
                                                 f'{current_term.name} [{current_term.identifier.value}] '
                                                 f'and its ancestor '
                                                 f'{ancestor_term.name} [{ancestor_term.identifier.value}]'))

        return ValidationResults(results)


class PhenotypicAbnormalityValidator(BaseOntologyRuleValidator):
    """
    Validator for checking that the term is a phenotypic abnormality
    (a descendant of Phenotypic abnormality HP:0000118).

    The validator replaces obsolete term IDs with the current term IDs before performing the validation.
    """

    def __init__(self, ontology: MinimalOntology):
        super().__init__(ontology)

    def validate(self, items: typing.Sequence[Identified]) -> ValidationResults:
        results = []
        for term in items:
            term_id = self._primary_term_id(term.identifier)
            ancestors = get_ancestors(self._ontology, source=term_id, include_source=False)
            if PHENOTYPIC_ABNORMALITY not in ancestors:
                term = self._ontology.get_term(term_id)
                results.append(
                    ValidationResult(
                        level=ValidationLevel.ERROR,
                        category='phenotypic_abnormality_descendant',
                        message=f'{term.name} [{term.identifier.value}] '
                                f'is not a descendant of Phenotypic abnormality [{PHENOTYPIC_ABNORMALITY.value}]'
                    )
                )

        return ValidationResults(results)


class ObsoleteTermIdsValidator(BaseOntologyRuleValidator):

    def __init__(self, ontology: MinimalOntology):
        super().__init__(ontology)

    def validate(self, items: typing.Sequence[Identified]) -> ValidationResults:
        results = []
        for term in items:
            current_id = self._primary_term_id(term.identifier)
            if current_id != term.identifier:
                current_term = self._ontology.get_term(term.identifier)
                results.append(
                    ValidationResult(
                        level=ValidationLevel.WARNING,
                        category='obsolete_term_id_is_used',
                        message=f'Using the obsolete {term.identifier.value} instead of {current_id.value} '
                                f'for {current_term.name}'
                    )
                )

        return ValidationResults(results)



import abc
import enum
import typing

from ._base import Identified, Named
from ._term_id import TermId


class MinimalTerm(Identified, Named, metaclass=abc.ABCMeta):
    """
    Minimal information regarding an ontology concept.

    `MinimalTerm` has an `identifier`, a `name`, a sequence of alternate IDs (IDs of obsolete terms that should be
    replaced by this term), and obsoletion status.
    """

    @staticmethod
    def create_minimal_term(term_id: TermId,
                            name: str,
                            alt_term_ids: typing.Sequence[TermId],
                            is_obsolete: bool):
        return DefaultMinimalTerm(term_id, name, alt_term_ids, is_obsolete)

    @property
    @abc.abstractmethod
    def alt_term_ids(self) -> typing.Sequence[TermId]:
        """
        :return: a sequence of identifiers of `Term`s that have been obsolete and replaced by this `Term`.
        """
        pass

    @property
    def is_current(self) -> bool:
        """
        :return: `True` if the term is current (*not* obsolete).
        """
        return not self.is_obsolete

    @property
    @abc.abstractmethod
    def is_obsolete(self) -> bool:
        """
        :return: `True` if the term has been obsolete.
        """
        pass

    def __eq__(self, other):
        return isinstance(other, MinimalTerm) \
            and self.identifier == other.identifier \
            and self.name == other.name \
            and self.alt_term_ids == other.alt_term_ids \
            and self.is_obsolete == other.is_obsolete

    def __str__(self):
        return f'MinimalTerm(' \
               f'identifier="{self.identifier}", ' \
               f'name="{self.name}", ' \
               f'alt_term_ids={self.alt_term_ids}, ' \
               f'is_obsolete={self.is_obsolete})'


class SynonymCategory(enum.Enum):
    """
    Synonym category.
    """
    EXACT = enum.auto()
    RELATED = enum.auto()
    BROAD = enum.auto()
    NARROW = enum.auto()


class SynonymType(enum.Enum):
    """
    Synonym type, as provided by Obographs.
    """
    LAYPERSON_TERM = enum.auto()
    ABBREVIATION = enum.auto()
    UK_SPELLING = enum.auto()
    OBSOLETE_SYNONYM = enum.auto()
    PLURAL_FORM = enum.auto()
    ALLELIC_REQUIREMENT = enum.auto()

    # We may not need these, unless we support ECTO
    # IUPAC_NAME = enum.auto()
    # INN = enum.auto()
    # BRAND_NAME = enum.auto()
    # IN_PART = enum.auto()
    # SYNONYM = enum.auto()
    # BLAST_NAME = enum.auto()
    # GENBANK_COMMON_NAME = enum.auto()
    # COMMON_NAME = enum.auto()

    def is_obsolete(self):
        return self == SynonymType.OBSOLETE_SYNONYM

    def is_current(self):
        return not self.is_obsolete()


class Synonym(Named):

    def __init__(self, name: str,
                 synonym_category: typing.Optional[SynonymCategory],
                 synonym_type: typing.Optional[SynonymType],
                 xrefs: typing.Optional[typing.Sequence[TermId]] = None):
        self._name = name
        self._scat = synonym_category
        self._stype = synonym_type
        self._xrefs = xrefs

    @property
    def name(self) -> str:
        return self._name

    @property
    def category(self) -> typing.Optional[SynonymCategory]:
        """
        :return: an instance of :class:`SynonymCategory` or ``None``.
        """
        return self._scat

    @property
    def synonym_type(self) -> typing.Optional[SynonymType]:
        """
        :return: an instance of :class:`SynonymType` or ``None``.
        """
        return self._stype

    @property
    def xrefs(self) -> typing.Optional[typing.Sequence[TermId]]:
        return self._xrefs

    def __eq__(self, other):
        isinstance(other, Synonym) \
        and self.name == other.name \
        and self.category == other.category \
        and self.synonym_type == other.synonym_type \
        and self.xrefs == other.xrefs

    def __str__(self):
        return f'Synonym(' \
               f'name="{self.name}", ' \
               f'category={self.category}, ' \
               f'synonym_type={self.synonym_type}, ' \
               f'xrefs="{self.xrefs}")'

    def __repr__(self):
        return str(self)


class Term(MinimalTerm, metaclass=abc.ABCMeta):
    """
    A representation of an ontology concept.
    """

    # TODO - add the remaining attributes from phenol's Term?

    @staticmethod
    def create_term(identifier: TermId,
                    name: str,
                    alt_term_ids: typing.Sequence[TermId],
                    is_obsolete: typing.Optional[bool],
                    definition: typing.Optional[str],
                    comment: typing.Optional[str],
                    synonyms: typing.Optional[typing.Sequence[Synonym]],
                    xrefs: typing.Optional[typing.Sequence[TermId]]):
        return DefaultTerm(identifier, name, alt_term_ids, is_obsolete, definition, comment, synonyms, xrefs)

    @property
    @abc.abstractmethod
    def definition(self) -> typing.Optional[str]:
        """
        :return: the term's definition.
        """
        pass

    @property
    @abc.abstractmethod
    def comment(self) -> str:
        """
        :return: the term's comment string.
        """
        pass

    @property
    @abc.abstractmethod
    def synonyms(self) -> typing.Optional[typing.Sequence[Synonym]]:
        """
        :return: all synonyms (including obsolete) of the term.
        """
        pass

    def current_synonyms(self) -> typing.Iterable[Synonym]:
        """
        :return: iterable over current (non-obsolete) synonyms of the term.
        """
        return self._synonyms_iter(lambda synonym:
                                   synonym.synonym_type is None
                                   or synonym.synonym_type.is_current())

    def obsolete_synonyms(self) -> typing.Iterable[Synonym]:
        """
        :return: iterable over obsolete synonyms of the term.
        """
        return self._synonyms_iter(lambda synonym:
                                   synonym.synonym_type is not None
                                   and synonym.synonym_type.is_obsolete())

    def _synonyms_iter(self, filter_f):
        if self.synonyms is None:
            return ()
        else:
            for synonym in self.synonyms:
                if filter_f(synonym):
                    yield synonym

    @property
    @abc.abstractmethod
    def xrefs(self) -> typing.Optional[typing.Sequence[TermId]]:
        """
        :return: term's cross-references
        """
        pass

    def __eq__(self, other):
        return MinimalTerm.__eq__(self, other) \
            and isinstance(other, Term) \
            and self.definition == other.definition \
            and self.comment == other.comment \
            and self.synonyms == other.synonyms \
            and self.xrefs == other.xrefs

    def __str__(self):
        return f'Term(' \
               f'identifier={self.identifier}, ' \
               f'name="{self.name}", ' \
               f'definition={self.definition}, ' \
               f'comment={self.comment}, ' \
               f'synonyms={self.synonyms}, ' \
               f'xrefs={self.xrefs}, ' \
               f'is_obsolete={self.is_obsolete}, ' \
               f'alt_term_ids="{self.alt_term_ids}")'


class DefaultMinimalTerm(MinimalTerm):

    def __init__(self, identifier: TermId,
                 name: str,
                 alt_term_ids: typing.Sequence[TermId],
                 is_obsolete: bool):
        self._id = identifier
        self._name = name
        self._alts = alt_term_ids
        self._is_obsolete = is_obsolete

    @property
    def identifier(self):
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def alt_term_ids(self) -> typing.Sequence[TermId]:
        return self._alts

    @property
    def is_obsolete(self) -> bool:
        return self._is_obsolete

    def __repr__(self):
        return f'DefaultMinimalTerm(' \
               f'identifier={self._id},' \
               f' name="{self._name}",' \
               f' is_obsolete={self._is_obsolete},' \
               f' alt_term_ids="{self._alts}")'


class DefaultTerm(Term):

    def __init__(self, identifier: TermId,
                 name: str,
                 alt_term_ids: typing.Sequence[TermId],
                 is_obsolete: typing.Optional[bool],
                 definition: typing.Optional[str],
                 comment: typing.Optional[str],
                 synonyms: typing.Optional[typing.Sequence[Synonym]],
                 xrefs: typing.Optional[typing.Sequence[TermId]]):
        self._id = identifier
        self._name = name
        self._alt_term_ids = alt_term_ids
        self._is_obsolete = False if is_obsolete is None else is_obsolete
        self._definition = definition
        self._comment = comment
        self._synonyms = synonyms
        self._xrefs = xrefs

    @property
    def identifier(self) -> TermId:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def alt_term_ids(self) -> typing.Sequence[TermId]:
        return self._alt_term_ids

    @property
    def is_obsolete(self) -> bool:
        return self._is_obsolete

    @property
    def definition(self) -> str:
        return self._definition

    @property
    def comment(self) -> str:
        return self._comment

    @property
    def synonyms(self) -> typing.Optional[typing.Sequence[Synonym]]:
        return self._synonyms

    @property
    def xrefs(self) -> typing.Optional[typing.Sequence[TermId]]:
        return self._xrefs

    def __repr__(self):
        return f'DefaultTerm(' \
               f'identifier={self._id}, ' \
               f'name="{self._name}", ' \
               f'definition={self._definition}, ' \
               f'comment={self._comment}, ' \
               f'synonyms={self._synonyms}, ' \
               f'xrefs={self._xrefs}, ' \
               f'is_obsolete={self._is_obsolete}, ' \
               f'alt_term_ids="{self._alt_term_ids}")'



import os
import unittest

import ddt
from pkg_resources import resource_filename

import hpotk as hp
from hpotk.model import TermId
from hpotk.ontology import MinimalOntology
from hpotk.ontology.load.obographs import load_minimal_ontology

TOY_HPO = resource_filename(__name__, os.path.join('data', 'hp.toy.json'))

hp.util.setup_logging()

HPO: MinimalOntology = load_minimal_ontology(TOY_HPO)


@ddt.ddt
class TestTraversal(unittest.TestCase):

    @ddt.data(
        ("HP:0001166", False, {"HP:0001238", "HP:0100807"}),
        ("HP:0001166", True, {"HP:0001166", "HP:0001238", "HP:0100807"})
    )
    @ddt.unpack
    def test_get_parents(self, source: str, include_source, expected):
        parents = hp.algorithm.get_parents(HPO, source, include_source)
        self.assertIsInstance(parents, frozenset)
        self.assertSetEqual(parents, {TermId.from_curie(val) for val in expected})

    @ddt.data(
        ("HP:0001166", False, {
            "HP:0001238", "HP:0100807", "HP:0011842", "HP:0033127", "HP:0002813", "HP:0040068",
            "HP:0002817", "HP:0011844", "HP:0011297", "HP:0000001", "HP:0001155", "HP:0040064",
            "HP:0000924", "HP:0000118", "HP:0001167"
        }),
        ("HP:0001166", True, {
            "HP:0001166", "HP:0001238", "HP:0100807", "HP:0011842", "HP:0033127", "HP:0002813", "HP:0040068",
            "HP:0002817", "HP:0011844", "HP:0011297", "HP:0000001", "HP:0001155", "HP:0040064", "HP:0000924",
            "HP:0000118", "HP:0001167"
        })
    )
    @ddt.unpack
    def test_get_ancestors(self, source: str, include_source, expected):
        ancestors = hp.algorithm.get_ancestors(HPO, source, include_source)
        self.assertIsInstance(ancestors, frozenset)
        self.assertSetEqual(ancestors, {TermId.from_curie(val) for val in expected})

    @ddt.data(
        ("HP:0001167", False, {"HP:0001238", "HP:0100807"}),
        ("HP:0001167", True, {"HP:0001167", "HP:0001238", "HP:0100807"})
    )
    @ddt.unpack
    def test_get_children(self, source: str, include_source, expected):
        children = hp.algorithm.get_children(HPO, source, include_source)
        self.assertIsInstance(children, frozenset)
        self.assertSetEqual(children, {TermId.from_curie(val) for val in expected})

    @ddt.data(
        ("HP:0001167", False, {"HP:0001166", "HP:0001238", "HP:0100807"}),
        ("HP:0001167", True, {"HP:0001167", "HP:0001166", "HP:0001238", "HP:0100807"})
    )
    @ddt.unpack
    def test_get_descendents(self, source: str, include_source, expected):
        descendents = hp.algorithm.get_descendents(HPO, source, include_source)
        self.assertIsInstance(descendents, frozenset)
        self.assertSetEqual(descendents, {TermId.from_curie(val) for val in expected})

    def test_we_get_correct_number_of_descendents(self):
        all_term_id = "HP:0000001"
        self.assertEqual(len(HPO) - 1, len(hp.algorithm.get_descendents(HPO, all_term_id)))
        self.assertEqual(len(HPO), len(hp.algorithm.get_descendents(HPO, all_term_id, include_source=True)))



from hpotk.model import Versioned
import warnings

# TODO[v0.3.0] - remove the deprecated file
warnings.warn('hpotk.ontology._attrs has been deprecated. Use hpotk.model.Versioned instead',
              DeprecationWarning, stacklevel=2)

